//+------------------------------------------------------------------+
//|                                                   TrailingCT.mqh |
//|                   Copyright 2009-2013, MetaQuotes Software Corp. |
//|                                              http://www.mql5.com |
//+------------------------------------------------------------------+
#include <Math\Stat\Math.mqh>
#include <Expert\ExpertTrailing.mqh>
#include <ct_9.mqh>
// wizard description start
//+------------------------------------------------------------------+
//| Description of the class                                         |
//| Title=Trailing Stop based on 'Category Theory' monoid-action concepts  |
//| Type=Trailing                                                    |
//| Name=CategoryTheory                                              |
//| ShortName=CT                                                     |
//| Class=CTrailingCT                                                |
//| Page=trailing_ct                                                 |
//| Parameter=Step,double,0.5,Trailing Step                          |
//| Parameter=LookbackOperation,int,0, Lookback Operation            |
//| Parameter=LookbackIdentity,int,0, Lookback Identity              |
//| Parameter=LookbackIdentityAct,int,0, Lookback Identity Action    |
//| Parameter=TimeframeOperation,int,0, Timeframe Operation          |
//| Parameter=TimeframeIdentity,int,0, Timeframe Identity            |
//| Parameter=TimeframeIdentityAct,int,0, Timeframe Identity Action  |
//| Parameter=AppliedpriceOperation,int,0, Appliedprice Operation    |
//| Parameter=AppliedpriceIdentity,int,0, Appliedprice Identity      |
//| Parameter=AppliedpriceIdentityAct,int,0, Appliedprice Identity Action|
//| Parameter=IndicatorOperation,int,0, Indicator Operation          |
//| Parameter=IndicatorIdentity,int,0, Indicator Identity            |
//| Parameter=DecisionOperation,int,0, Decision Operation            |
//| Parameter=DecisionIdentity,int,0, Decision Identity              |
//| Parameter=CorrelationInverted,bool,false, Correlation Inverted   |
//+------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| Class CTrailingCT.                                               |
//| Appointment: Class traling stops with 'Category Theory'          |
//|               monoid-action concepts.                            |
//| Derives from class CExpertTrailing.                              |
//+------------------------------------------------------------------+
int                        __LOOKBACKS[8] = {1,2,3,4,5,6,7,8};
ENUM_TIMEFRAMES            __TIMEFRAMES[8] = {PERIOD_H1,PERIOD_H2,PERIOD_H3,PERIOD_H4,PERIOD_H6,PERIOD_H8,PERIOD_H12,PERIOD_D1};
ENUM_APPLIED_PRICE         __APPLIEDPRICES[4] = { PRICE_MEDIAN, PRICE_TYPICAL, PRICE_OPEN, PRICE_CLOSE };
string                     __INDICATORS[2] = { "RSI", "BOLLINGER_BANDS" };
string                     __DECISIONS[2] = { "TREND", "RANGE" };

#define                    __CORR 5
int                        __LOOKBACKS_A[10] = {1,2,3,4,5,6,7,8,9,10};
ENUM_TIMEFRAMES            __TIMEFRAMES_A[10] = {PERIOD_H1,PERIOD_H2,PERIOD_H3,PERIOD_H4,PERIOD_H6,PERIOD_H8,PERIOD_H12,PERIOD_D1,PERIOD_W1,PERIOD_MN1};
ENUM_APPLIED_PRICE         __APPLIEDPRICES_A[5] = { PRICE_MEDIAN, PRICE_TYPICAL, PRICE_OPEN, PRICE_CLOSE, PRICE_WEIGHTED };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CTrailingCT : public CExpertTrailing
  {
protected:
   //--- adjusted parameters
   double m_step;                // trailing step
   
   CMonoid<double>               m_lookback;
   EOperations                   m_lookback_operation;
   int                           m_lookback_identity;
   
   CMonoid<double>               m_timeframe;
   EOperations                   m_timeframe_operation;
   int                           m_timeframe_identity;
   
   CMonoid<double>               m_appliedprice;
   EOperations                   m_appliedprice_operation;
   int                           m_appliedprice_identity;
   
   CMonoid<double>               m_indicator;
   EOperations                   m_indicator_operation;
   int                           m_indicator_identity;
   
   CMonoid<double>               m_decision;
   EOperations                   m_decision_operation;
   int                           m_decision_identity;
   
   //
   
   CMonoidAction<double,double>  m_lookback_act;
   CMonoidAction<double,double>  m_timeframe_act;
   CMonoidAction<double,double>  m_appliedprice_act;
   
   bool                          m_correlation_inverted;
   
   int                           m_lookback_identity_act;
   int                           m_timeframe_identity_act;
   int                           m_appliedprice_identity_act;
   
   int                           m_source_size;       // Source Size

public:
   //--- methods of setting adjustable parameters
   
   
   
   //--- method of verification of settings
   virtual bool         ValidationSettings(void);
   //--- method of creating the indicator and timeseries
   virtual bool         InitIndicators(CIndicators *indicators);
   //--- methods of checking if the market models are formed
   virtual bool         CheckTrailingStopLong(CPositionInfo *position,double &sl,double &tp);
   virtual bool         CheckTrailingStopShort(CPositionInfo *position,double &sl,double &tp);
   //---
                        CTrailingCT(void);
                        ~CTrailingCT(void);
   //--- methods of initialization of protected data
   void                 Step(double value)                  { m_step=value;                                 }
   
   void                 LookbackOperation(int value)        { m_lookback_operation=(EOperations)value;      }
   void                 LookbackIdentity(int value)         { m_lookback_identity=value;                    }
   
   void                 TimeframeOperation(int value)       { m_timeframe_operation=(EOperations)value;     }
   void                 TimeframeIdentity(int value)        { m_timeframe_identity=value;                   }
   
   void                 AppliedpriceOperation(int value)    { m_appliedprice_operation=(EOperations)value;  }
   void                 AppliedpriceIdentity(int value)     { m_appliedprice_identity=value;                }
   
   void                 IndicatorOperation(int value)       { m_indicator_operation=(EOperations)value;     }
   void                 IndicatorIdentity(int value)        { m_indicator_identity=value;                   }
   
   void                 DecisionOperation(int value)        { m_decision_operation=(EOperations)value;      }
   void                 DecisionIdentity(int value)         { m_decision_identity=value;                    }
   
   void                 CorrelationInverted(bool value)     { m_correlation_inverted=value;                 }
   
   void                 LookbackIdentityAct(int value)      { m_lookback_identity_act=value;                }
   void                 TimeframeIdentityAct(int value)     { m_timeframe_identity_act=value;               }
   void                 AppliedpriceIdentityAct(int value)  { m_appliedprice_identity_act=value;            }

protected:
   
   CElement<double>     m_element;

   int                  GetLookback(CMonoid<double> &M,int &L[]);
   ENUM_TIMEFRAMES      GetTimeframe(CMonoid<double> &M,ENUM_TIMEFRAMES &T[]);
   ENUM_APPLIED_PRICE   GetAppliedprice(CMonoid<double> &M,ENUM_APPLIED_PRICE &A[],int Lookback,ENUM_TIMEFRAMES Timeframe);
   double               GetIndicator(int Lookback,ENUM_TIMEFRAMES Timeframe,ENUM_APPLIED_PRICE Appliedprice);
   
   int                  GetDecision(int Lookback,double Indicator);

   int                  GetLookback(int Lookback,int &L[]){ int _l=-1; for(int i=0;i<8;i++){ if(L[i]==Lookback){ _l=i; break; }}  return(_l); }
   int                  GetTimeframe(ENUM_TIMEFRAMES Timeframe,ENUM_TIMEFRAMES &T[]){ int _t=-1; for(int i=0;i<8;i++){ if(T[i]==Timeframe){ _t=i; break; }}  return(_t); }
   int                  GetAppliedprice(ENUM_APPLIED_PRICE Appliedprice,ENUM_APPLIED_PRICE &A[]){ int _a=-1; for(int i=0;i<4;i++){ if(A[i]==Appliedprice){ _a=i; break; }}  return(_a); }
   
   double               m_lookbacks[__CORR];
   double               m_timeframes[__CORR];
   double               m_appliedprices[__CORR];
   
   double               m_priceranges[__CORR];
   
   void                 Operate(CMonoid<double> &M,EOperations &O,int &OutputIndex);
   
   void                 GetAction(int &LookBack,ENUM_TIMEFRAMES &TimeFrame,ENUM_APPLIED_PRICE &AppliedPrice);
   
  };
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
CTrailingCT::CTrailingCT(void)
  {
//--- initialization of protected data
      m_used_series=USE_SERIES_TIME+USE_SERIES_SPREAD+USE_SERIES_HIGH+USE_SERIES_LOW+USE_SERIES_CLOSE;
      
      m_lookback.Cardinality(8);
      m_lookback.Operation(m_lookback_operation);
      m_lookback.Identity(m_lookback_identity);
      
      m_timeframe.Cardinality(8);
      m_timeframe.Operation(m_timeframe_operation);
      m_timeframe.Identity(m_timeframe_identity);
      
      m_appliedprice.Cardinality(4);
      m_appliedprice.Operation(m_appliedprice_operation);
      m_appliedprice.Identity(m_appliedprice_identity);
      
      m_indicator.Cardinality(2);
      m_indicator.Operation(m_indicator_operation);
      m_indicator.Identity(m_indicator_identity);
      
      m_decision.Cardinality(2);
      m_decision.Operation(m_decision_operation);
      m_decision.Identity(m_decision_identity);
      
      m_step=0.5;
      
      ArrayInitialize(m_lookbacks,0.0);
      ArrayInitialize(m_timeframes,0.0);
      ArrayInitialize(m_appliedprices,0.0);
      
      ArrayInitialize(m_priceranges,0.0);
      
      m_lookback_act.Cardinality(10);
      m_lookback_act.Operation(m_lookback_operation);
      m_lookback_act.Identity(m_lookback_identity_act);
      
      m_timeframe_act.Cardinality(10);
      m_timeframe_act.Operation(m_timeframe_operation);
      m_timeframe_act.Identity(m_timeframe_identity_act);
      
      m_appliedprice_act.Cardinality(5);
      m_appliedprice_act.Operation(m_appliedprice_operation);
      m_appliedprice_act.Identity(m_appliedprice_identity_act);
  }
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
CTrailingCT::~CTrailingCT(void)
  {
  }
//+------------------------------------------------------------------+
//| Validation settings protected data.                              |
//+------------------------------------------------------------------+
bool CTrailingCT::ValidationSettings(void)
  {
//--- validation settings of additional filters
   if(!CExpertTrailing::ValidationSettings())
      return(false);

   if(m_period!=PERIOD_H1){
      printf(__FUNCSIG__+" Timeframe should be one-hour! ");
      return(false);}
//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| Create indicators.                                               |
//+------------------------------------------------------------------+
bool CTrailingCT::InitIndicators(CIndicators *indicators)
  {
//--- check pointer
   if(indicators==NULL)
      return(false);
//--- initialization of indicators and timeseries of additional filters
   if(!CExpertTrailing::InitIndicators(indicators))
      return(false);
//--- 
//--- ok
   return(true);
  }
//+------------------------------------------------------------------+
//| Checking trailing stop and/or profit for long position.          |
//+------------------------------------------------------------------+
bool CTrailingCT::CheckTrailingStopLong(CPositionInfo *position,double &sl,double &tp)
  {
//--- check
      if(position==NULL)
         return(false);
      
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
      
      int _lookback=GetLookback(m_lookback,__LOOKBACKS);
      ENUM_TIMEFRAMES _timeframe=GetTimeframe(m_timeframe,__TIMEFRAMES);
      ENUM_APPLIED_PRICE _appliedprice=GetAppliedprice(m_appliedprice,__APPLIEDPRICES,_lookback,_timeframe);
      //
      GetAction(_lookback,_timeframe,_appliedprice);
      //
      double _indicator=GetIndicator(_lookback,_timeframe,_appliedprice);
      int _decision=GetDecision(_lookback,_indicator);
      
      double _type=_decision/100.0;
   
      double _atr=fmax(2.0*m_spread.GetData(_x)*m_symbol.Point(),m_high.GetData(_x)-m_low.GetData(_x))*(_type);
      
      double _sl=m_low.GetData(_x)-(m_step*_atr);
      
      double level =NormalizeDouble(m_symbol.Bid()-m_symbol.StopsLevel()*m_symbol.Point(),m_symbol.Digits());
      double new_sl=NormalizeDouble(_sl,m_symbol.Digits());
      double pos_sl=position.StopLoss();
      double base  =(pos_sl==0.0) ? position.PriceOpen() : pos_sl;
      
      sl=EMPTY_VALUE;
      tp=EMPTY_VALUE;
      if(new_sl>base && new_sl<level)
         sl=new_sl;
//---
   return(sl!=EMPTY_VALUE);
  }
//+------------------------------------------------------------------+
//| Checking trailing stop and/or profit for short position.         |
//+------------------------------------------------------------------+
bool CTrailingCT::CheckTrailingStopShort(CPositionInfo *position,double &sl,double &tp)
  {
//--- check
      if(position==NULL)
         return(false);
      
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
      
      int _lookback=GetLookback(m_lookback,__LOOKBACKS);
      ENUM_TIMEFRAMES _timeframe=GetTimeframe(m_timeframe,__TIMEFRAMES);
      ENUM_APPLIED_PRICE _appliedprice=GetAppliedprice(m_appliedprice,__APPLIEDPRICES,_lookback,_timeframe);
      //
      GetAction(_lookback,_timeframe,_appliedprice);
      //
      double _indicator=GetIndicator(_lookback,_timeframe,_appliedprice);
      int _decision=GetDecision(_lookback,_indicator);
      
      double _type=_decision/100.0;
   
      double _atr=fmax(2.0*m_spread.GetData(_x)*m_symbol.Point(),m_high.GetData(_x)-m_low.GetData(_x))*(_type);
      
      double _sl=m_high.GetData(_x)+(m_step*_atr);
      
      double level =NormalizeDouble(m_symbol.Ask()+m_symbol.StopsLevel()*m_symbol.Point(),m_symbol.Digits());
      double new_sl=NormalizeDouble(_sl,m_symbol.Digits());
      double pos_sl=position.StopLoss();
      double base  =(pos_sl==0.0) ? position.PriceOpen() : pos_sl;
      
      sl=EMPTY_VALUE;
      tp=EMPTY_VALUE;
      if(new_sl<base && new_sl>level)
         sl=new_sl;
//---
      return(sl!=EMPTY_VALUE);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::GetAction(int &LookBack,ENUM_TIMEFRAMES &TimeFrame,ENUM_APPLIED_PRICE &AppliedPrice)
   {
      m_high.Refresh(-1);
      m_low.Refresh(-1);
      
      int _x=StartIndex();
      //
      double _value=0.0;m_element.Let();
      //
      int _lookback_index=GetLookback(LookBack,__LOOKBACKS_A);
      if(m_lookback.Get(_lookback_index,m_element) && m_element.Get(0,_value))
      {
         for(int c=__CORR-1;c>0;c--){ m_lookbacks[c]=m_lookbacks[c-1]; } m_lookbacks[0]=_value;
      }
      //
      _value=0.0;m_element.Let();
      int _timeframe_index=GetTimeframe(TimeFrame,__TIMEFRAMES_A);
      if(m_timeframe.Get(_timeframe_index,m_element) && m_element.Get(0,_value))
      {
         for(int c=__CORR-1;c>0;c--){ m_timeframes[c]=m_timeframes[c-1]; } m_timeframes[0]=_value;
      }
      //
      _value=0.0;m_element.Let();
      int _appliedprice_index=GetAppliedprice(AppliedPrice,__APPLIEDPRICES_A);
      if(m_appliedprice.Get(_appliedprice_index,m_element) && m_element.Get(0,_value))
      {
         for(int c=__CORR-1;c>0;c--){ m_priceranges[c]=m_priceranges[c-1]; } m_appliedprices[0]=_value;
      }
      //
      for(int c=__CORR-1;c>0;c--){ m_appliedprices[c]=m_appliedprices[c-1]; } m_priceranges[0]=(m_high.GetData(_x)-m_low.GetData(_x))-(m_high.GetData(_x+1)-m_low.GetData(_x+1));
      //
      double _lookback_r=0.0;MathCorrelationSpearman(m_lookbacks,m_priceranges,_lookback_r);
      double _timeframe_r=0.0;MathCorrelationSpearman(m_timeframes,m_priceranges,_timeframe_r);
      double _appliedprice_r=0.0;MathCorrelationSpearman(m_appliedprices,m_priceranges,_appliedprice_r);
      
      if(!m_correlation_inverted)
      {
         if(_lookback_r>=_timeframe_r && _lookback_r>=_appliedprice_r)
         {
            LookBack=GetLookback(m_lookback_act,__LOOKBACKS_A);
         }
         else if(_timeframe_r>=_lookback_r && _timeframe_r>=_appliedprice_r)
         {
            TimeFrame=GetTimeframe(m_timeframe_act,__TIMEFRAMES_A);
         }
         else if(_appliedprice_r>=_lookback_r && _appliedprice_r>=_timeframe_r)
         {
            AppliedPrice=GetAppliedprice(m_appliedprice_act,__APPLIEDPRICES_A,LookBack,TimeFrame);
         }
      }
      else
      {
         if(_lookback_r<=_timeframe_r && _lookback_r<=_appliedprice_r)
         {
            LookBack=GetLookback(m_lookback_act,__LOOKBACKS_A);
         }
         else if(_timeframe_r<=_lookback_r && _timeframe_r<=_appliedprice_r)
         {
            TimeFrame=GetTimeframe(m_timeframe_act,__TIMEFRAMES_A);
         }
         else if(_appliedprice_r<=_lookback_r && _appliedprice_r<=_timeframe_r)
         {
            AppliedPrice=GetAppliedprice(m_appliedprice_act,__APPLIEDPRICES_A,LookBack,TimeFrame);
         }
      }
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CTrailingCT::Operate(CMonoid<double> &M,EOperations &O,int &OutputIndex)
   {
      OutputIndex=-1;
      //
      double _values[];
      ArrayResize(_values,M.Cardinality());ArrayInitialize(_values,0.0);
      //
      for(int i=0;i<M.Cardinality();i++)
      {
         m_element.Let();
         if(M.Get(i,m_element))
         {
            if(!m_element.Get(0,_values[i]))
            {
               printf(__FUNCSIG__+" Failed to get double for 1 at: "+IntegerToString(i+1));
            }
         }
         else{ printf(__FUNCSIG__+" Failed to get element for 1 at: "+IntegerToString(i+1)); }
      }
      
      //
      
      if(O==OP_LEAST)
      {
         OutputIndex=0;
         double _least=_values[0];
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_least>_values[i]){ _least=_values[i]; OutputIndex=i; }
         }
      }
      else if(O==OP_MOST)
      {
         OutputIndex=0;
         double _most=_values[0];
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_most<_values[i]){ _most=_values[i]; OutputIndex=i; }
         }
      }
      else if(O==OP_CLOSEST)
      {
         double _mean=0.0;
         for(int i=0;i<M.Cardinality();i++)
         {
            _mean+=_values[i];
         }
         _mean/=M.Cardinality();
         
         OutputIndex=0;
         double _closest=fabs(_values[0]-_mean);
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_closest>fabs(_values[i]-_mean)){ _closest=fabs(_values[i]-_mean); OutputIndex=i; }
         }
      }
      else if(O==OP_FURTHEST)
      {
         double _mean=0.0;
         for(int i=0;i<M.Cardinality();i++)
         {
            _mean+=_values[i];
         }
         _mean/=M.Cardinality();
         
         OutputIndex=0;
         double _furthest=fabs(_values[0]-_mean);
         for(int i=0;i<M.Cardinality();i++)
         {
            if(_furthest<fabs(_values[i]-_mean)){ _furthest=fabs(_values[i]-_mean); OutputIndex=i; }
         }
      }
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int CTrailingCT::GetLookback(CMonoid<double> &M,int &L[])
   {
      m_close.Refresh(-1);
      
      int _x=StartIndex();
      
      for(int i=0;i<M.Cardinality();i++)
      {
         int _period=(L[i]*PeriodSeconds(PERIOD_H4))/PeriodSeconds(m_period);
         double _value=fabs(m_close.GetData(_x)-m_close.GetData(_x+_period))/(fabs(m_close.GetData(_x)-m_close.GetData(_x+_period))+fabs(m_close.GetData(_x+_period)-m_close.GetData(_x+_period+_period)));
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!M.Set(i,m_element,true))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
   
      int _i_out=-1;
      //
      Operate(M,m_lookback_operation,_i_out);
      
      if(_i_out==-1){ return(4); }
      
      return(4*L[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
ENUM_TIMEFRAMES CTrailingCT::GetTimeframe(CMonoid<double> &M, ENUM_TIMEFRAMES &T[])
   {
      for(int i=0;i<M.Cardinality();i++)
      {
         ResetLastError();
         double _value=0.0;
         double _buffer[];ArrayResize(_buffer,3);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
         if(CopyClose(m_symbol.Name(),T[i],0,3,_buffer)>=3)
         {
            _value=fabs(_buffer[0]-_buffer[1])/(fabs(_buffer[0]-_buffer[1])+fabs(_buffer[1]-_buffer[2]));
         }
         else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(T[i])+" close prices. err: "+IntegerToString(GetLastError())); }
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!M.Set(i,m_element,true))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for lookback. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      //
      Operate(M,m_timeframe_operation,_i_out);
      
      if(_i_out==-1){ return(INVALID_HANDLE); }
      
      return(T[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
ENUM_APPLIED_PRICE CTrailingCT::GetAppliedprice(CMonoid<double> &M,ENUM_APPLIED_PRICE &A[],int Lookback,ENUM_TIMEFRAMES Timeframe)
   {
      for(int i=0;i<M.Cardinality();i++)
      {
         ResetLastError();
         double _value=0.0;
         if(A[i]==PRICE_CLOSE)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(m_symbol.Name(),Timeframe,0,Lookback,_buffer)>=Lookback)
            {
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(A[i]==PRICE_OPEN)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyOpen(m_symbol.Name(),Timeframe,0,Lookback,_buffer)>=Lookback)
            {
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" open prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(A[i]==PRICE_MEDIAN)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            if(CopyHigh(m_symbol.Name(),Timeframe,0,Lookback,_h)>=Lookback && CopyLow(m_symbol.Name(),Timeframe,0,Lookback,_l)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=0.5*(_h[ii]+_l[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(A[i]==PRICE_TYPICAL)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            double _c[];ArrayResize(_c,Lookback);ArrayInitialize(_c,0.0);
            if(CopyHigh(m_symbol.Name(),Timeframe,0,Lookback,_h)>=Lookback && CopyLow(m_symbol.Name(),Timeframe,0,Lookback,_l)>=Lookback && CopyClose(m_symbol.Name(),Timeframe,0,Lookback,_c)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=(1.0/3.0)*(_h[ii]+_l[ii]+_c[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(A[i]==PRICE_WEIGHTED)
         {
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            double _h[];ArrayResize(_h,Lookback);ArrayInitialize(_h,0.0);
            double _l[];ArrayResize(_l,Lookback);ArrayInitialize(_l,0.0);
            double _c[];ArrayResize(_c,Lookback);ArrayInitialize(_c,0.0);
            if(CopyHigh(m_symbol.Name(),Timeframe,0,Lookback,_h)>=Lookback && CopyLow(m_symbol.Name(),Timeframe,0,Lookback,_l)>=Lookback && CopyClose(m_symbol.Name(),Timeframe,0,Lookback,_c)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _buffer[ii]=(1.0/4.0)*(_h[ii]+_l[ii]+_c[ii]+_c[ii]);
               }
               _value=MathStandardDeviation(_buffer);
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" high and/or low prices. err: "+IntegerToString(GetLastError())); }
         }
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!M.Set(i,m_element,true))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for applied price. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      //
      Operate(M,m_appliedprice_operation,_i_out);
      
      if(_i_out==-1){ return(4); }
      
      return(A[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double CTrailingCT::GetIndicator(int Lookback,ENUM_TIMEFRAMES Timeframe,ENUM_APPLIED_PRICE Appliedprice)
   {
      m_spread.Refresh(-1);
      
      int _x=StartIndex();
      
      double _indicator[2];
      ArrayInitialize(_indicator,0.0);
      
      for(int i=0;i<2;i++)
      {
         ResetLastError();
         double _value=0.0;
         //
         if(__INDICATORS[i]=="RSI")
         {
            double _spread=(m_symbol.Point()*m_spread.GetData(_x));
            double _current_gain=_spread,_current_loss=_spread;
            double _gain=_spread,_loss=_spread;
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(m_symbol.Name(),Timeframe,0,Lookback,_buffer)>=Lookback)
            {
               if(_buffer[0]>_buffer[1]){ _current_gain+=_buffer[0]-_buffer[1]; }
               else if(_buffer[1]>_buffer[0]){ _current_loss+=_buffer[1]-_buffer[0]; }
               
               for(int ii=1;ii<Lookback-1;ii++)
               {
                  if(_buffer[ii]>_buffer[ii+1]){ _gain+=_buffer[ii]-_buffer[ii+1]; }
                  else if(_buffer[ii+1]>_buffer[ii]){ _loss+=_buffer[ii+1]-_buffer[ii]; }
               }
               
               _gain/=(Lookback-1);
               _loss/=(Lookback-1);
               
               _gain+=_current_gain;
               _loss+=_current_loss;
               
               _value=(_gain/_loss);
               _value+=1.0;
               _value=100.0/_value;
               _value=100.0-_value;
               
               _indicator[0]=_value;
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         else if(__INDICATORS[i]=="BOLLINGER_BANDS")
         {
            double _base=0.0,_upper=0.0,_lower=0.0;
            double _buffer[];ArrayResize(_buffer,Lookback);ArrayInitialize(_buffer,0.0);ArraySetAsSeries(_buffer,true);
            if(CopyClose(m_symbol.Name(),Timeframe,0,Lookback,_buffer)>=Lookback)
            {
               for(int ii=0;ii<Lookback;ii++)
               {
                  _base+=_buffer[ii];
               }
               
               _base/=(Lookback);
               
               double _deviation=MathStandardDeviation(_buffer);
               
               _upper+=(_base+(2.0*_deviation));
               _lower+=(_base-(2.0*_deviation));
               
               _value=((1.0+((_buffer[0]-_base)/(fabs(_buffer[0]-_base)+(_upper-_lower))))*50.0);
               
               _indicator[1]=_value;
            }
            else{ printf(__FUNCSIG__+" Failed to copy: "+EnumToString(Timeframe)+" close prices. err: "+IntegerToString(GetLastError())); }
         }
         
         //
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!m_indicator.Set(i,m_element,true))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for indicator. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      Operate(m_indicator,m_indicator_operation,_i_out);
      
      return(_indicator[_i_out]);
   }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int CTrailingCT::GetDecision(int Lookback,double Indicator)
   {
      m_close.Refresh(-1);
      
      int _x=StartIndex();
      
      for(int i=0;i<2;i++)
      {
         ResetLastError();
         double _value=0.0;
         //
         int _max_index=-1,_min_index=-1;
         double _max=m_close.MaxValue(_x,Lookback,_max_index);
         double _min=m_close.MinValue(_x,Lookback,_min_index);
         
         if(__DECISIONS[i]=="TREND")
         {
            int _trends=0;
            if(m_close.GetData(_x)>m_close.GetData(_x+Lookback))
            {
               for(int ii=_x;ii<_x+Lookback;ii++)
               {
                  if(m_close.GetData(ii)>m_close.GetData(ii+1))
                  {
                     _trends++;
                     _value+=(m_close.GetData(ii)-m_close.GetData(ii+1));
                  }
               }
            }
            else if(m_close.GetData(_x)<m_close.GetData(_x+Lookback))
            {
               for(int ii=_x;ii<_x+Lookback;ii++)
               {
                  if(m_close.GetData(ii)<m_close.GetData(ii+1))
                  {
                     _trends++;
                     _value+=(m_close.GetData(ii+1)-m_close.GetData(ii));
                  }
               }
            }
            
            if(_trends>0){ _value/=_trends; }
            if(_max-_min>0.0){ _value/=(_max-_min); }
         }
         else if(__DECISIONS[i]=="RANGE")
         {
            int _ranges=0;
            if(m_close.GetData(_x)>m_close.GetData(_x+Lookback))
            {
               for(int ii=_x;ii<_x+Lookback;ii++)
               {
                  if(m_close.GetData(ii)<m_close.GetData(ii+1))
                  {
                     _ranges++;
                     _value+=(m_close.GetData(ii+1)-m_close.GetData(ii));
                  }
               }
            }
            else if(m_close.GetData(_x)<m_close.GetData(_x+Lookback))
            {
               for(int ii=_x;ii<_x+Lookback;ii++)
               {
                  if(m_close.GetData(ii)>m_close.GetData(ii+1))
                  {
                     _ranges++;
                     _value+=(m_close.GetData(ii)-m_close.GetData(ii+1));
                  }
               }
            }
            
            if(_ranges>0){ _value/=_ranges; }
            if(_max-_min>0.0){ _value/=(_max-_min); }
         }
         
         //
         
         m_element.Let();
         m_element.Cardinality(1);
         if(m_element.Set(0,_value))
         {
            ResetLastError();
            if(!m_decision.Set(i,m_element,true))
            {
               printf(__FUNCSIG__+" Failed to assign element at index: "+IntegerToString(i)+", for decision. ERR: "+IntegerToString(GetLastError()));
            }
         }
      }
      
      int _i_out=-1;
      Operate(m_decision,m_decision_operation,_i_out);
      
      double _decision=0.0;
      
      if(_i_out==0)
      {
         if(m_close.GetData(_x)>m_close.GetData(_x+Lookback))
         {
            _decision=Indicator;
         }
         else if(m_close.GetData(_x)<m_close.GetData(_x+Lookback))
         {
            _decision=100.0-Indicator;
         }
      }
      else if(_i_out==1)
      {
         if(m_close.GetData(_x)>m_close.GetData(_x+Lookback))
         {
            _decision=100.0-Indicator;
         }
         else if(m_close.GetData(_x)<m_close.GetData(_x+Lookback))
         {
            _decision=Indicator;
         }
      }
      
      return(int(round(_decision)));
   }
//+------------------------------------------------------------------+
